import argparse
import sys
import traceback
import subprocess as sb
import json
import os
import re
from hashlib import md5
import itertools
from datetime import datetime as dt
from pathlib import Path


class State:
    def __init__(self, name, payload):
        self.payload = payload
        self.merge_cached_content = args.merge_cached_content
        self.path = self._get_location(name)

    @staticmethod
    def remove(name):
        path = State._get_location(name)
        if path.exists():
            path.unlink()
            return path

    @staticmethod
    def _resources_dir():
        if sys.platform.startswith("linux"):
            return Path("/opt/microsoft/mdatp/resources")

        if sys.platform.startswith("darwin"):
            return Path("/Applications/Microsoft Defender ATP.app/Contents/MacOS/wdavdaemon_enterprise.app/Contents/Resources")

        raise ValueError("platform not supported: {}".format(sys.platform))

    @staticmethod
    def _get_location(name):
        basedir = State._resources_dir() / "cache"
        if not basedir.exists():
            basedir.mkdir()
        return (basedir / name).with_suffix(".json")

    def __enter__(self):
        if not self.path.exists():
            return

        if not self.merge_cached_content:
            return

        with open(self.path, 'r') as f:
            old_payload = json.load(f)

        current_handlers = {h['name'] for h in self.payload['handlers']}

        for handler in old_payload['handlers']:
            name, pid = handler['name'], handler['pid']

            if name in current_handlers or not Path(name).exists():
                continue
            
            handler['active'] = False 
            self.payload['handlers'].append(handler)
            self.payload['environ'][pid] = old_payload['environ'][pid]
            
    def __exit__(self, *args, **kwargs):
        with open(self.path, 'w') as f:
            json.dump(self.payload, f, separators=(',', ':'))

def validate_state_id(dest):
    if not re.match('^[A-Za-z0-9._]+$', dest):
            raise ValueError("state id invalid: {}".format(dest))
    return dest

def validate_command_name(dest):
    commands = [x.strip() for x in dest.split(",")]
    for command in commands:
        if not re.match('^[A-Za-z0-9]+$', command):
            raise ValueError("command name invalid: {}".format(command))
    return commands

def split_by_comma(dest):
    return [x.strip() for x in dest.split(",")]

def validate_limit(dest):
    limit = int(dest)
    if limit < 1:
        raise ValueError("limit is too small")
    return limit

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--ScriptName", dest='script_name')
    parser.add_argument("--id", type=validate_state_id)
    parser.add_argument("--limit", type=validate_limit, default=100, help='limit results. defaults to 100.')
    parser.add_argument("--filter-command", type=validate_command_name, dest='lsof_commands', help='commands to filter')
    parser.add_argument("--filter-name", type=split_by_comma, help='filter handler name')
    parser.add_argument("--filter-env", type=split_by_comma, help='filter env vars')
    parser.add_argument("--merge-cached-content", action='store_true', default=True) 
    parser.add_argument("--remove", type=validate_state_id, help="remove a specific state, by id")

    args = parser.parse_args()
    if not args.remove and any([not args.id, not args.lsof_commands]):
        parser.error("you must pass --id and --filter-command")
    return args



def _run_cmd(cmd):
    with sb.Popen(cmd, stdout=sb.PIPE, bufsize=1, universal_newlines=True, shell=True) as p:
        yield from map(str.strip, p.stdout)

def grab_environ(pid):
    # use 'ps' to grab the env variables of given pid
    cmd = "ps eww -o command {} | tr ' ' '\n' | tail -n+3".format(int(pid))
    for line in _run_cmd(cmd):
        if not args.filter_env or any(x in line for x in args.filter_env):
            yield line

def take(n, l):
    for i, item in enumerate(l):
        if i > n:
            break
        yield item

def grab_handlers():
    # run lsof with given parameters, take only first N handlers (ignore the rest)
    cmd = 'lsof -a -c {} 2>/dev/null'.format(' -c '.join(args.lsof_commands))
    handlers_gen = _run_cmd(cmd)

    # if lsof returned output, the first line would be the column names
    column_names = next(handlers_gen, "").strip().lower().split()
    if not column_names:
        return

    # iterate each handler & parse it
    for handler_line in handlers_gen:
        ts = dt.utcnow().strftime('%Y-%m-%d %H:%M:%S UTC')
        handler = dict(zip(column_names, handler_line.split()),timestamp=ts, active=True)
        handler_name = handler.get('name')

        # handler name must exist
        if not handler_name:
            continue

        if not args.filter_name or any(x in handler_name for x in args.filter_name):
            yield handler

def main():
    if os.geteuid() != 0:
        raise ValueError("script needs to be run as root")

    if args.remove:
        path = State.remove(args.remove)
        return 0 if path is not None else 1

    handlers = list(take(args.limit, grab_handlers()))
    environ = {h['pid']: list(grab_environ(h['pid'])) for h in handlers}
    payload = dict(scriptVersion=1, cmdline=" ".join(sys.argv[1:]), handlers=handlers,environ=environ)

    with State(name=args.id, payload=payload):
        print(json.dumps(payload),file=sys.stdout)

    return 0

if __name__ == "__main__":
    try:
        # workaround: arguments are incorrectly passed to the script. this fixes the issue.
        sys.argv = list(itertools.chain(*[x.split(" ") for x in sys.argv]))
        args = parse_args()
        sys.exit(main())
    except SystemExit as se:
        sys.exit(se.code)
    except:
        print(traceback.format_exc(),file=sys.stderr)
        sys.exit(1)
